/*
 * bcamctl
 * Low level binary that talks to the camera over a serial port.
 *
 * (c) Copyright 2000 Free Software Foundation
 *
 * This software comes with ABSOLUTELY NO WARRANTY. Use it on your own risk.
 */
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#define STX 0x02
#define ETX 0x03
#define ACK 0x06
#define NAK 0x15

#define NAME "bcamctl"
#define SBSIZE     1200         /* Let it fit into an Ethernet frame */
#define VERSZ  100		/* normally only 4 characters... */

#define PROTO_TCP    6          /* == getprotobyname("tcp") */
/* #define DFL_PORT   5000  */  /* Annex-like TCP port */

int opendev(char *name);
int open_dev(char *name);
int open_annex(const char *name);
void command(int, int argc, char *argv[]);
void do_reset_index(int dfd, char arg[]);
void do_beep(int dfd, char arg1[], char arg2[]);
void do_echo(int);
void do_grab(int dfd, char arg1[], char arg2[]);
void do_get_index(int);
void do_inc_addr(int dfd, char arg[]);
void do_write(int dfd, char arg[]);
void do_idle(int);
void do_upload_thumbnail(int dfd, char arg[]);
void do_port(int dfd, char arg[]);
void do_set_video(int dfd, char arg[]);
void do_read_reg(int dfd, char arg[]);
void do_set_expband(int dfd, char arg1[], char arg2[]);
void do_upload_image(int dfd, char arg[]);
void do_get_version(int dfd);
void do_write_reg(int dfd, char arg1[], char arg2[]);
void do_get_expband(int dfd);
void do_grab_result(int dfd);
void do_quit_selftest(int dfd);
void upload_wrapper(int dfd, char arg[], char cmd, int len);
int cmdping(int dfd, const char *cmd, int len);
void cmdrrep(int dfd, char *buf, int len);
void chkbusy(const char *reply);
void Usage(void);
int read_all(int fd, char *data, int size);

int main(int argc, char *argv[]) {
	int dfd;

	/*
	 * Let's be friendly and not try to open anything
	 * if user just asks for -help.
	 */
	if (argc < 3 || argv[1][0] == '-') Usage();

	dfd = opendev(argv[1]);

	argc -= 2;
	argv += 2;
	command(dfd, argc, argv);

	close(dfd);
	exit(0);
	return 0;
}

/*
 * Open a physical device or a network connection.
 * We dictate that host argument has a colon in it
 * to prevent common case problems with missing arguments in scripts, etc.
 */
int opendev(char *name) {
	int dfd = -1;

	if (strchr(name, '/') != NULL) {		/* Device or FIFO */
		dfd = open_dev(name);
	} else if (strchr(name, ':') != NULL) {		/* host:port */
		dfd = open_annex(name);
	} else {
		Usage();
	}
	return dfd;
}

int open_dev(char *name) {
	int dfd;
	struct termios tb;

	if ((dfd = open(name, O_RDWR|O_NONBLOCK)) < 0) {
		fprintf(stderr, NAME ": Cannot open %s: %s\n", name,
		    strerror(errno));
		exit(1);
	}

	(void) ioctl(dfd, TCFLSH, TCIOFLUSH);

	memset(&tb, 0, sizeof(tb));
	tb.c_iflag = 0;
	tb.c_oflag = 0;
	tb.c_cflag = CS8 | CREAD | CLOCAL | B57600;	/* HUPCL? */
	tb.c_lflag = 0;
	tb.c_cc[VMIN] = 1;
	tb.c_cc[VTIME] = 0;
	if (ioctl(dfd, TCSETS, &tb) != 0) {
		fprintf(stderr, NAME ": Cannot set tty %s: %s\n", name,
		    strerror(errno));
		exit(1);
	}

	/* make the input blocking again */
	(void) fcntl(dfd, F_SETFL, fcntl(dfd, F_GETFL, 0) & ~O_NONBLOCK);

	return dfd;
}

int open_annex(const char *name) {
	static char hostbuf[64];
	int hnlen;
	char *pstr;
	int port;
	int dfd;
	struct hostent *hp;
	struct sockaddr_in anadr;

	pstr = strchr(name, ':');
	if (pstr == 0) abort();
	hnlen = pstr - name;
	pstr++;
	if (hnlen > 63) {
		fprintf(stderr, NAME ": host name too long (%.63s>)\n", name);
		exit(1);
	}
	memcpy(hostbuf, name, hnlen);
	hostbuf[hnlen] = 0;

	if (!isdigit(*pstr)) {		/* XXX Insufficient */
		fprintf(stderr, NAME ": bad port (%.10s)\n", pstr);
		exit(1);
	}
	port = atoi(pstr);

	if ((hp = gethostbyname(hostbuf)) == NULL) {
		fprintf(stderr, NAME ": cannot resolve %s\n", hostbuf);
		exit(1);
	}
	if (hp->h_addr == 0) {
		fprintf(stderr, NAME ": no address for %s\n", hostbuf);
		exit(1);
	}
	if (hp->h_addrtype != AF_INET) {  /* so that anadr won't overflow */
		fprintf(stderr, NAME ": no IP address for %s\n", hostbuf);
		exit(1);
	}
	anadr.sin_family = hp->h_addrtype;
	memcpy(&anadr.sin_addr, hp->h_addr, sizeof(anadr.sin_addr));
	anadr.sin_port = htons(port);

	if ((dfd = socket(AF_INET, SOCK_STREAM, PROTO_TCP)) < 0) {
		fprintf(stderr, NAME ": socket: %s\n", strerror(errno));
		exit(1);
	}

	if (connect(dfd, (struct sockaddr *)&anadr, sizeof(anadr)) < 0) {
		fprintf(stderr, NAME ": connect: %s\n", strerror(errno));
		exit(1);
	}

	return dfd;
}

/*
 * Switch to different commands according to the argument.
 */
void command(int dfd, int argc, char *argv[]) {

	switch (argv[0][0]) {
	case 'A':		/* A x        Reset image index */
		do_reset_index(dfd, argv[1]);
		break;
	case 'B':		/* B dur num  Beep(s) */
		do_beep(dfd, argv[1], argv[2]);
		break;
	case 'E':		/* E char     Echo */
		do_echo(dfd);
		break;
	case 'G':		/* G flag     Grab */
		do_grab(dfd, argv[1], argv[2]);
		break;
	case 'I':		/* I          Get image index */
		do_get_index(dfd);
		break;
	case 'J':		/* J inc      Increment address and read byte */
		do_inc_addr(dfd, argv[1]);
		break;
	case 'K':		/* K byte     Write byte */
		do_write(dfd, argv[1]);
		break;
	case 'L':		/* L          Go Idle */
		do_idle(dfd);
		break;
	case 'M':		/* M          Upload thumbnail */
		do_upload_thumbnail(dfd, argv[1]);
		break;
	case 'P':		/* P val      Write port, returns old port */
		do_port(dfd, argv[1]);
		break;
	case 'Q':		/* Q mode     Set video mode */
		do_set_video(dfd, argv[1]);
		break;
	case 'R':		/* R x        Read camreg */
		do_read_reg(dfd, argv[1]);
		break;
	case 'T':		/* T min max  Set expband */
		do_set_expband(dfd, argv[1], argv[2]);
		break;
	case 'U':		/* U file     Upload image */
		do_upload_image(dfd, argv[1]);
		break;
	case 'V':		/* V          Get firmware version */
		do_get_version(dfd);
		break;
	case 'W':		/* W x val    Write camreg */
		do_write_reg(dfd, argv[1], argv[2]);
		break;
	case 'X':		/* X          Get expband */
		do_get_expband(dfd);
		break;
	case 'Y':		/* Y          Grab result */
		do_grab_result(dfd);
		break;
	case 'Z':		/* Z          Quit selftest */
		do_quit_selftest(dfd);
		break;
	default:
		Usage();
	}
}

/*
 * You are in a maze of little functions, all alike.
 */

/*
 * Image counter can be 0 ... 6.
 * 6 means that all frames are used up.
 */
void do_reset_index(int dfd, char arg[]) {
	char cmd[2];
	char reply[4];
	int x;

	if (arg == NULL) {
		Usage();
	}
	if (!isdigit(arg[0])) {
		fprintf(stderr, NAME ": bad argument for 'A'\n");
		exit(1);
	}
	x = atoi(arg);
	if (x < 0 || x > 6) {
		fprintf(stderr, NAME ": bad vaule for 'A'\n");
		exit(1);
	}

	cmd[0] = 'A';
	cmd[1] = x;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'a') {
		fprintf(stderr, NAME ": 'A' mismatch\n");
		exit(1);
	}
}

/*
 * XXX Seems to work fine on my camera but does not produce any sound.
 * --zaitcev
 */
void do_beep(int dfd, char arg1[], char arg2[]) {
	char cmd[3];
	char reply[4];
	int dur, num;

	if (arg1 == NULL || arg2 == NULL) {
		Usage();
	}
	if (!isdigit(arg1[0]) || !isdigit(arg2[0])) {
		fprintf(stderr, NAME ": bad argument for 'B'\n");
		exit(1);
	}
	dur = atoi(arg1);
	num = atoi(arg2);
	if (dur < 0 || dur > 255 || num < 0 || num > 255) {
		fprintf(stderr, NAME ": bad vaule for 'B'\n");
		exit(1);
	}

	cmd[0] = 'B';
	cmd[1] = dur;
	cmd[1] = num;

	if (cmdping(dfd, cmd, 3) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'b') {
		fprintf(stderr, NAME ": 'B' mismatch\n");
		exit(1);
	}
}

/*
 */
void do_echo(int dfd) {
	char reply[4];

	if (cmdping(dfd, "E\151", 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'e' || reply[2] != '\151') {
		fprintf(stderr, NAME ": echo mismatch\n");
		exit(1);
	}
}

/*
 * This function ('G') starts a grab. Check result of the grab with ('Y').
 */
void do_grab(int dfd, char arg1[], char arg2[]) {
	char cmd[2];
	char reply[4];
	int delay, retries;

	if (arg1 == NULL || arg2 == NULL) {
		Usage();
	}
	if (!isdigit(arg1[0]) || !isdigit(arg2[0])) {
		fprintf(stderr, NAME ": bad argument for 'G'\n");
		exit(1);
	}
	delay = atoi(arg1);
	retries = atoi(arg2);
	if (delay < 0 || delay > 15 || retries < 0 || retries > 7) {
		fprintf(stderr, NAME ": bad vaule for 'G'\n");
		exit(1);
	}

	/*
	 * Actual retry table:
	 *    Retry value:         0   1   2   3   4   5   6   7
	 *    Number of retries:   0  16  32  48  64  80  96 112
	 * Bit 0x80 is reserved for "grab anyways".
	 */
	cmd[0] = 'G';
	cmd[1] = (retries << 4) | delay;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'g') {
		fprintf(stderr, NAME ": 'G' mismatch\n");
		exit(1);
	}
}

/*
 */
void do_get_index(int dfd) {
	char reply[4];
	int index;

	if (cmdping(dfd, "I\0", 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'i') {
		fprintf(stderr, NAME ": get_index mismatch\n");
		exit(1);
	}
	if ((index = reply[2] & 0xFF) > 6) {
		fprintf(stderr, NAME ": got bad index %d\n", index);
		exit(1);
	}

	printf("%d\n", index);
}

/*
 * Increment, then get byte.
 */
void do_inc_addr(int dfd, char arg[]) {
	char cmd[3];
	char reply[5];
	int x;

	if (arg == NULL) {
		Usage();
	}
	if (!isdigit(arg[0])) {
		fprintf(stderr, NAME ": bad argument for 'J'\n");
		exit(1);
	}
	x = atoi(arg);
	if (x < 0 || x > 32767) {
		fprintf(stderr, NAME ": bad vaule for 'J'\n");
		exit(1);
	}

	cmd[0] = 'J';
	cmd[1] = x;
	cmd[1] = x >> 8;

	if (cmdping(dfd, cmd, 3) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 5);
	chkbusy(reply);
	if (reply[1] != 'j') {
		fprintf(stderr, NAME ": increment mismatch\n");
		exit(1);
	}

	printf("%d\n", reply[3] & 0xFF);
}

/*
 * Write byte. Camera autoincrements.
 */
void do_write(int dfd, char arg[]) {
	char cmd[2];
	char reply[4];
	int x;

	if (arg == NULL) {
		Usage();
	}
	if (!isdigit(arg[0])) {
		fprintf(stderr, NAME ": bad argument for 'K'\n");
		exit(1);
	}
	x = atoi(arg);
	if (x < 0 || x > 255) {
		fprintf(stderr, NAME ": bad vaule for 'K'\n");
		exit(1);
	}

	cmd[0] = 'K';
	cmd[1] = x;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'k') {
		fprintf(stderr, NAME ": write protocol mismatch\n");
		exit(1);
	}
}

/*
 * Note: Camera does not sleep yet when the response is returned.
 */
void do_idle(int dfd) {
	char reply[4];

	if (cmdping(dfd, "L\0", 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'l' || reply[2] != '\0') {
		fprintf(stderr, NAME ": idle mismatch\n");
		exit(1);
	}
}

/*
 */
void do_upload_thumbnail(int dfd, char arg[]) {
	/* 1240 bytes = 40 columns * (0 blank + 31 visible) */
	upload_wrapper(dfd, arg, 'M', 1240);
}

/*
 * XXX Do an operation which is called "PORT0".
 */
void do_port(int dfd, char arg[]) {
	char cmd[2];
	char reply[4];
	int x;

	if (arg == NULL) Usage();
	if (!isdigit(arg[0])) {
		fprintf(stderr, NAME ": bad argument for 'P'\n");
		exit(1);
	}
	x = atoi(arg);
	if (x < 0 || x > 255) {
		fprintf(stderr, NAME ": bad vaule for 'P'\n");
		exit(1);
	}

	cmd[0] = 'P';
	cmd[1] = x;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'p') {
		fprintf(stderr, NAME ": port protocol mismatch\n");
		exit(1);
	}

	printf("%d\n", reply[2] & 0xFF);
}

/*
 * Enable/disable continuus video taking, which cannot be used
 * outside of the manufacturing testbed, alas.
 */
void do_set_video(int dfd, char arg[]) {
	fprintf(stderr, "'Q' is not supported because it's not useful\n");
}

/*
 * Read camreg.
 */
void do_read_reg(int dfd, char arg[]) {
	char cmd[2];
	char reply[4];
	int x;

	if (arg == NULL) Usage();
	if (!isdigit(arg[0])) {
		fprintf(stderr, NAME ": bad argument for 'R'\n");
		exit(1);
	}
	x = atoi(arg);
	if (x < 0 || x > 255) {
		fprintf(stderr, NAME ": bad vaule for 'R'\n");
		exit(1);
	}

	cmd[0] = 'R';
	cmd[1] = x;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'r') {
		fprintf(stderr, NAME ": port protocol mismatch\n");
		exit(1);
	}

	printf("%d\n", reply[2] & 0xFF);
}

/*
 * Set exposure thresholds.
 */
void do_set_expband(int dfd, char arg1[], char arg2[]) {
	char cmd[3];
	char reply[4];
	int min, max;

	if (arg1 == NULL || arg2 == NULL) {
		Usage();
	}
	if (!isdigit(arg1[0]) || !isdigit(arg2[0])) {
		fprintf(stderr, NAME ": bad argument for 'T'\n");
		exit(1);
	}
	min = atoi(arg1);
	max = atoi(arg2);
	if (min < 0 || min > 255 || max < 0 || max > 255) {
		fprintf(stderr, NAME ": bad value for 'T'\n");
		exit(1);
	}
	if (min >= max) {
		fprintf(stderr, NAME ": min must be less than max\n");
		exit(1);
	}

	cmd[0] = 'T';
	cmd[1] = min;
	cmd[2] = max;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 't') {
		fprintf(stderr, NAME ": 'T' mismatch\n");
		exit(1);
	}

	if (reply[2] != 0) {
		fprintf(stderr, NAME ": 'T' error %d\n", reply[2] & 0xFF);
		exit(1);
	}
}

/*
 * Upload full image.
 */
void do_upload_image(int dfd, char arg[]) {
	/* 20664 bytes = 164 col * (2 blank + 124 visible lines) */
	upload_wrapper(dfd, arg, 'U', 20664);
}

/*
 * Get version (variable size).
 */
void do_get_version(int dfd) {
	char reply[VERSZ+2];
	int len;
	int rc;

	if (cmdping(dfd, "V\0", 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}

	cmdrrep(dfd, reply, 3);
	chkbusy(reply);
	if (reply[1] != 'v') {
		fprintf(stderr, NAME ": version reply mismatch\n");
		exit(1);
	}
	if ((len = reply[2]) < 2) {
		fprintf(stderr, NAME ": error gettin version\n");
		exit(1);
	}
	/* They say length includes the trailing null. */

	if ((rc = read_all(dfd, reply, len+1)) < 0) {
		fprintf(stderr, NAME ": read: %s\n", strerror(errno));
		exit(1);
	}
	if (rc < len+1) {
		fprintf(stderr, NAME ": short version read\n");
		exit(1);
	}

	reply[len-1] = 0;
	printf("%s\n", reply);
}

/*
 * Write a camera register.
 */
void do_write_reg(int dfd, char arg1[], char arg2[]) {
	char cmd[3];
	char reply[4];
	int addr, val;

	if (arg1 == NULL || arg2 == NULL) {
		Usage();
	}
	if (!isdigit(arg1[0]) || !isdigit(arg2[0])) {
		fprintf(stderr, NAME ": bad argument for 'W'\n");
		exit(1);
	}
	addr = atoi(arg1);
	val = atoi(arg2);
	if (addr < 0 || addr > 255 || val < 0 || val > 255) {
		fprintf(stderr, NAME ": bad value for 'W'\n");
		exit(1);
	}

	cmd[0] = 'W';
	cmd[1] = addr;
	cmd[2] = val;

	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'w') {
		fprintf(stderr, NAME ": 'W' mismatch\n");
		exit(1);
	}
	if (reply[2] != 0) {
		fprintf(stderr, NAME ": 'W' error %d\n", reply[2] & 0xFF);
		exit(1);
	}
}

/*
 */
void do_get_expband(int dfd) {
	char reply[4];

	if (cmdping(dfd, "X\0", 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'x') {
		fprintf(stderr, NAME ": get_expband mismatch\n");
		exit(1);
	}

	printf("min %d max %d\n", reply[2] & 0xFF, reply[3] & 0xFF);
}

/*
 * Query a result of the last grab.
 * While grab is in progress, a busy error will be returned as usual.
 */
void do_grab_result(int dfd) {
	char reply[4];
	int res;

	if (cmdping(dfd, "Y\0", 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 4);
	chkbusy(reply);
	if (reply[1] != 'y') {
		fprintf(stderr, NAME ": get_expband mismatch\n");
		exit(1);
	}

	res = reply[2] & 0xFF;
	switch (res) {
	case 0:
		printf("0 OK\n");
		break;
	case 0x80:
		printf("0x80 Memory is full\n");
		break;
	case 0x81:
		printf("0x81 Cannot attain exposure\n");
		break;
	default:
		printf("0x%x Unknown result\n", res);
	}
	printf("min %d max %d\n", reply[2] & 0xFF, reply[3] & 0xFF);
}

/*
 */
void do_quit_selftest(int dfd) {
	fprintf(stderr, "'Z' not implemented because camera must quit "
	    "selftest by itself\n");
	exit(1);
}

/*
 * Helpers for uploading.
 * "uploading" means "uploading to the host".
 */
void upload_helper(int dfd, int ffd, char cmdc, int elen) {
	char *buf;
	char cmd[2];
	char reply[6];
	int dlen;
	int rc;
	int i;

	cmd[0] = cmdc;
	cmd[1] = 0;
	if (cmdping(dfd, cmd, 2) != 0) {
		fprintf(stderr, NAME ": cmdping protocol error\n");
		exit(1);
	}
	cmdrrep(dfd, reply, 6);
	chkbusy(reply);
	if (reply[1] != tolower(cmdc)) {
		fprintf(stderr, NAME ": upload reply mismatch\n");
		exit(1);
	}
	dlen = (reply[2] & 0xFF) * ((reply[3] + reply[4]) & 0xFF);
	if (elen != dlen) {
		fprintf(stderr, NAME ": device returned unexpected"
		    " image size %d, N1=%d N2=%d N3=%d, expected %d\n",
		    dlen, reply[2] & 0xFF, reply[3] & 0xFF, reply[4] & 0xFF,
		    elen);
		exit(1);
	}
	if ((reply[5] & 0xFF) != 16) {
		fprintf(stderr, NAME ": device returned unexpected"
		    " N4=%d, expected %d\n", reply[5] & 0xFF, 16);
		exit(1);
	}

	if ((buf = malloc(dlen + 17)) == NULL) {
		fprintf(stderr, NAME ": no core\n");
		exit(1);
	}

	if ((rc = read_all(dfd, buf, dlen+17)) < 0) {
		fprintf(stderr, NAME ": read: %s\n", strerror(errno));
		exit(1);
	}
	if (rc < dlen+17) {
		fprintf(stderr, NAME ": short thumbnail read\n");
		exit(1);
	}
	if (buf[dlen+16] != ETX) {
		fprintf(stderr, NAME ": missing ETX after image (0x%x)\n",
		    buf[dlen+16] & 0xFF);
		exit(1);
	}

	if ((rc = write(ffd, buf, dlen)) < 0) {
		fprintf(stderr, NAME ": write: %s\n", strerror(errno));
		exit(1);
	}
	if (rc < dlen) {
		fprintf(stderr, NAME ": short image write\n");
		exit(1);
	}

	printf("status");
	for (i = 0; i < 16; i++) {
		printf(" 0x%02x", buf[dlen + i] & 0xFF);
	}
	printf("\n");

	free(buf);
}

void upload_wrapper(int dfd, char arg[], char cmd, int len) {
	int ffd;

	if (arg == NULL) Usage();

	if (arg[0] == '-') {
		upload_helper(dfd, 1, cmd, len);
	} else {
		if ((ffd = open(arg, O_WRONLY|O_CREAT, 0644)) < 0) {
			fprintf(stderr, NAME ": Cannot open output: %s\n",
			    strerror(errno));
			exit(1);
		}
		upload_helper(dfd, ffd, cmd, len);
		if (close(ffd) != 0) {
			fprintf(stderr, NAME ": Cannot close output: %s\n",
			    strerror(errno));
			exit(1);  /* Heck, nothing we can do */
		}
	}
}

/*
 * Test if the response is "Busy".
 */
void chkbusy(const char *reply) {
	if (reply[0] != STX) {
		fprintf(stderr, NAME ": no STX in reply\n");
		exit(1);
	}
	if (reply[1] == '!') {
		switch (reply[2]) {
		case 0x20:
			fprintf(stderr, NAME ": busy, wait function\n");
			break;
		case 0x21:
			fprintf(stderr, NAME ": busy, exposure\n");
			break;
		case 0x22:
			fprintf(stderr, NAME ": busy, continuous video\n");
			break;
		case 0x23:
			fprintf(stderr, NAME ": busy, raising shutter\n");
			break;
		case 0x24:
			fprintf(stderr, NAME ": busy, remote command\n");
			break;
		case 0x25:
			fprintf(stderr, NAME ": busy, resetting counter\n");
			break;
		default:
			fprintf(stderr, NAME ": busy, code %d\n",
			    reply[2] & 0xFF);
		}
		exit(1);
	}
}

/*
 * Send a command to the camera, get ACK or NAK.
 * Later replies to the command are too diverse to be handled
 * by a command-agnostic function like this.
 *
 * System errors cause a message and a bail out.
 *
 * Note that we do not time out reading not here not anywhere else.
 *
 * XXX Remove printouts - library function.
 */
int cmdping(int dfd, const char *cmd, int len) {
	char buff[5];
	char *p;
	int i;
	int rc;

	if (len < 0 || len > 3) return -1;
	p = buff;
	*p++ = STX;
	for (i = 0; i < len; i++) *p++ = cmd[i];
	*p++ = ETX;

	if (write(dfd, buff, p-buff) < 0) goto error;
	if ((rc = read(dfd, buff, 1)) < 0) goto error;
	if (rc == 0) return -1;

	if (buff[0] != ACK) return -1;

	return 0;

error:
	fprintf(stderr, NAME ": cmdping: %s\n", strerror(errno));
	exit(1);
	return -1;
}

/*
 * Since our overall model is so simple and lets us bail out
 * whenever we want, we use it to have this helper function.
 * It reduces the visual clutter in do_xxx().
 */
void cmdrrep(int dfd, char *buf, int len) {
	int rc;

	if ((rc = read_all(dfd, buf, len)) < 0) {
		fprintf(stderr, NAME ": cmdrep: %s\n", strerror(errno));
		exit(1);
	}
	if (rc < len) {
		fprintf(stderr, NAME ": short read on reply\n");
		exit(1);
	}
}

void Usage(void) {
	fprintf(stderr,
	    "Usage: " NAME " {/dev/tty<port>|host:5000} <command>\n");
	fprintf(stderr, "Commands and arguments are:\n");
	fprintf(stderr, "A x        Reset image index\n");
	fprintf(stderr, "B dur num  Beep(s)\n");
	fprintf(stderr, "E char     Echo\n");
	fprintf(stderr, "G flag     Grab\n");
	fprintf(stderr, "I          Get image index\n");
	fprintf(stderr, "J inc      Increment address and read byte\n");
	fprintf(stderr, "K byte     Write byte\n");
	fprintf(stderr, "L          Go Idle\n");
	fprintf(stderr, "M          Upload thumbnail\n");
	fprintf(stderr, "P val      Write port\n");
	fprintf(stderr, "Q mode     Set video mode\n");
	fprintf(stderr, "R x        Read camreg\n");
	fprintf(stderr, "T min max  Set expband\n");
	fprintf(stderr, "U file     Upload image\n");
	fprintf(stderr, "V          Get version\n");
	fprintf(stderr, "W x val    Write camreg\n");
	fprintf(stderr, "X          Get expband\n");
	fprintf(stderr, "Y          Grab result\n");
	fprintf(stderr, "Z          Quit selftest\n");
	exit(1);
}

/*
 * Read bytes until the requested size is satisfied.
 * Normally this is done when reading from a network socket,
 * but we must to it while reading our serial connection too
 * because we set VMIN=1. If we set it differently, some broken
 * setups stop working (reading off a terminal server, etc.).
 */
int read_all(int fd, char *data, int size) {
	int togo;
	int rc;

	togo = size;
	while (togo > 0) {
		if ((rc = read(fd, data, togo)) < 0) return -1;
		if (rc == 0) {
			return size - togo;
		}
		togo -= rc;
		data += rc;
	}
	return size;
}
